
#include "worker.h"
#include "stats.h"

#include "kr2/kord/kord_server.h"
#include "kr2/kord/robot_model.h"
#include "kr2/kord/robot_controller.h"
#include "kr2/kord/robot_control_action.h"


#include "kr2/kord/protocol/KORDFrames.h"
#include "kr2/kord/protocol/StatusFrameBuilder.h"
#include "kr2/kord/protocol/ContentFrameBuilder.h"
#include "kr2/kord/protocol/ContentFrameParser.h"
#include "kr2/kord/protocol/DataFormatDescription.h"

#include "kr2_rc_api/internals/console.h"

#include <mutex>
#include <chrono>

#include <boost/crc.hpp>
typedef std::chrono::time_point<std::chrono::high_resolution_clock> timepoint;

namespace kkp = kr2::kord::protocol;

// TODO:
struct CaptureFRM {
    unsigned int frame_id;

    uint32_t fail_to_read_empty = 0;
    uint32_t fail_to_read_error = 0;

    // storing capture frame raw data
    std::chrono::_V2::system_clock::time_point recent_captured_timestamp;
    timespec recent_captured_content_timestamp;

    size_t   max_frames_in_tick;
    int64_t  faulty_frames_start;

    kkp::EKORDItemID         iid;
    kkp::KORDFrame           kord_frame;
    kkp::ContentFrameParser  content_parser;
    kkp::ContentFrameBuilder content_builder;

    timepoint rcv_cmd_joints_ts;
    kkp::DataFormatDescription dfd_cmd_move_J = kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eCommandMoveJ);
    kkp::ContentItem move_J;

    timepoint rcv_cmd_linear_ts;
    kkp::DataFormatDescription dfd_cmd_move_L = kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eCommandMoveL);
    kkp::ContentItem move_L;

    timepoint arm_status_req_ts;
    kkp::DataFormatDescription dfd_arm_status_req = kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eRequestStatusArm);
    kkp::ContentItem status_req;

    timepoint rcv_cmd_jfw_ts;
    kkp::DataFormatDescription dfd_cmd_jfw =  kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eCommandJointFirmware);
    kkp::ContentItem cmd_jfw;

    timepoint rcv_req_sys_ts;
    kkp::DataFormatDescription dfd_req_system =  kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eRequestSystem);
    kkp::ContentItem req_system;

    timepoint rcv_req_transf_ts;
    kkp::DataFormatDescription dfd_req_transfer =  kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eRequestTransfer);
    kkp::ContentItem req_transfer;

    timepoint rcv_req_io_ts;
    kkp::DataFormatDescription dfd_req_io =  kkp::DataFormatDescription::makeItemDescriptionLatest(kkp::EKORDItemID::eCommandSetIODigitalOut);
    kkp::ContentItem req_io;

    timepoint rcv_cmd_direct_ts;
    kkp::DataFormatDescription dfd_cmd_move_D = kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eCommandMoveDirect);
    kkp::ContentItem move_D;

    timepoint rcv_cmd_velocity_ts;
    kkp::DataFormatDescription dfd_cmd_move_V = kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eCommandMoveVelocity);
    kkp::ContentItem move_V;

    timepoint rcv_cmd_setFrame_ts;
    kkp::DataFormatDescription dfd_cmd_set_Frame = kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eCommandSetFrame);
    kkp::ContentItem set_Frame;

    timepoint rcv_cmd_setLoad_ts;
    kkp::DataFormatDescription dfd_cmd_set_Load = kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eCommandSetLoad);
    kkp::ContentItem set_Load;

    timepoint rcv_cmd_cleanAlarm_ts;
    kkp::DataFormatDescription dfd_cmd_clean_Alarm = kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eCommandCleanAlarm);
    kkp::ContentItem clean_Alarm;

    timepoint rcv_resp_status_ts;
    kkp::DataFormatDescription dfd_resp_status =  kkp::DataFormatDescription::makeItemDescriptionLatest (kkp::EKORDItemID::eStatusEchoResponse);
    kkp::ContentItem resp_status;

    boost::crc_optimal<16, 0x8005, 0xFFFF, 0, true, true> crc;

    void resetCaptureFrames()
    {
        using namespace std::chrono_literals;

        memset(&kord_frame, 0x00, sizeof (kkp::KORDFrame));
        rcv_cmd_joints_ts = timepoint(0ms);
        rcv_cmd_linear_ts = timepoint(0ms);
        arm_status_req_ts = timepoint(0ms);
        rcv_cmd_jfw_ts    = timepoint(0ms);
        rcv_req_sys_ts    = timepoint(0ms);
        rcv_req_transf_ts = timepoint(0ms);
        rcv_resp_status_ts= timepoint(0ms);
    }

    bool checkCRC(const uint16_t& a_received_crc, kkp::ContentItem &item){
        crc.reset(0xFFFF);
        crc.process_bytes(item.getItemData(), item.getItemDataLength() - 2);
        if (crc.checksum() != a_received_crc){
            CLOG_ERR("CRC doesn't match: " << crc.checksum() << " , " << a_received_crc);
            return false;
        }
        else{
            return true;
            //CLOG_INFO("CRC does match: " << crc.checksum() << " , " << received_crc);
        }
    }

    void captureIIDInfo(const kkp::EKORDItemID &iid, kkp::ContentItem &item)
    {
        switch (iid)
            {
                case kkp::EKORDItemID::eCommandMoveJ:
                {
                    // CRC check
                    if (!checkCRC(item.getData<uint16_t> (dfd_cmd_move_J.getOffset(kkp::EKORDDataID::eCRCValue)), item)){
                        return; //packet drop
                    }
                    
                    this->move_J            = item;
                    this->frame_id          = ntohs(this->kord_frame.frame_id_);
                    this->rcv_cmd_joints_ts = std::chrono::high_resolution_clock::now();
                    break;
                }
                case kkp::EKORDItemID::eCommandMoveL:
                {
                    // CRC check
                    if (!checkCRC(item.getData<uint16_t> (dfd_cmd_move_L.getOffset(kkp::EKORDDataID::eCRCValue)), item)){
                        return; // packet drop
                    }

                    this->move_L            = item;
                    this->frame_id          = ntohs(this->kord_frame.frame_id_);
                    this->rcv_cmd_linear_ts = std::chrono::high_resolution_clock::now();
                    break;
                }
                case kkp::EKORDItemID::eRequestStatusArm:
                {
                    this->status_req        = item;
                    this->arm_status_req_ts = std::chrono::high_resolution_clock::now();
                    break;
                }
                case kkp::EKORDItemID::eCommandJointFirmware:{
                    this->cmd_jfw           = item;
                    this->rcv_cmd_jfw_ts    = std::chrono::high_resolution_clock::now();
                    break;
                }
                case kkp::EKORDItemID::eStatusEchoResponse:
                {   
                    //CLOG_INFO("[KORD CAPTURE] Status Echo captured");
                    this->resp_status        = item;
                    this->rcv_resp_status_ts = std::chrono::high_resolution_clock::now();
                    break;
                }
                
                case kkp::EKORDItemID::eRequestSystem:
                {   
                    CLOG_INFO("[KORD CAPTURE] Request system captured");
                    this->req_system        = item;
                    this->rcv_req_sys_ts    = std::chrono::high_resolution_clock::now();
                    break;
                }
                case kkp::EKORDItemID::eRequestTransfer:
                {   
                    CLOG_INFO("[KORD CAPTURE] Request Transfer captured");
                    this->req_transfer           = item;
                    this->rcv_req_transf_ts    = std::chrono::high_resolution_clock::now();
                    break;
                }
                case kkp::EKORDItemID::eCommandSetIODigitalOut:
                {
                    CLOG_INFO("[KORD CAPTURE] Request IOSet digital captured");
                    this->req_io           = item;
                    this->rcv_req_io_ts    = std::chrono::high_resolution_clock::now();
                    break;
                }
                case kkp::EKORDItemID::eCommandMoveDirect:
                {
                    // CRC check
                    if (!checkCRC(item.getData<uint16_t> (dfd_cmd_move_D.getOffset(kkp::EKORDDataID::eCRCValue)), item)){
                        return; // packet drop
                    }
                    CLOG_INFO("[KORD CAPTURE] Request Direct Move captured");
                    this->move_D            = item;
                    this->frame_id          = ntohs(this->kord_frame.frame_id_);
                    this->rcv_cmd_direct_ts = std::chrono::high_resolution_clock::now();
                    break;
                }

                case kkp::EKORDItemID::eCommandMoveVelocity:
                {
                    // CRC check
                    if (!checkCRC(item.getData<uint16_t> (dfd_cmd_move_V.getOffset(kkp::EKORDDataID::eCRCValue)), item)){
                        return; // packet drop
                    }
                    CLOG_INFO("[KORD CAPTURE] Request Velocity Move captured");
                    this->move_V            = item;
                    this->frame_id          = ntohs(this->kord_frame.frame_id_);
                    this->rcv_cmd_velocity_ts = std::chrono::high_resolution_clock::now();
                    break;
                }

                case kkp::EKORDItemID::eCommandSetFrame:
                {
                    CLOG_INFO("[KORD CAPTURE] Request Set Frame captured");
                    this->set_Frame            = item;
                    this->frame_id          = ntohs(this->kord_frame.frame_id_);
                    this->rcv_cmd_setFrame_ts = std::chrono::high_resolution_clock::now();
                    break;
                }

                case kkp::EKORDItemID::eCommandSetLoad:
                {
                    CLOG_INFO("[KORD CAPTURE] Request Set Load captured");
                    this->set_Load            = item;
                    this->frame_id          = ntohs(this->kord_frame.frame_id_);
                    this->rcv_cmd_setLoad_ts = std::chrono::high_resolution_clock::now();
                    break;
                }
                
                case kkp::EKORDItemID::eCommandCleanAlarm:
                {
                    CLOG_INFO("[KORD CAPTURE] Request Clean Alarm captured");
                    this->clean_Alarm            = item;
                    this->frame_id          = ntohs(this->kord_frame.frame_id_);
                    this->rcv_cmd_cleanAlarm_ts = std::chrono::high_resolution_clock::now();
                    break;
                }

                default:
                    break;
            }
    }

};

struct TransmitFRM {
    kkp::KORDFrame kord_frame;

    std::vector<uint8_t> buffer;

    unsigned int seq_num_ = 0;
    kkp::StatusFrameBuilder status_builder;

    kkp::ContentFrameBuilder content_builder;

    void resetTransmitFrames()
    {
        memset(&kord_frame, 0x00, sizeof (kkp::KORDFrame));
        status_builder.clear();
        content_builder.clear();
    }

};


class Counter : public Worker {

    // Do we need the Builder interface?

    public:

    private:
        /// implement run for your specific Worker
		void run(){ }
};
